//
// Poscorr3xmm.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to correct a source list for an offset in positions
//              calculated using the ELLBETA PSF
//
//#include <ctime>

#include "CalServer.h"
#include "Dal.h"

#include "Poscorr3xmm.h"
#include "PositionUtils.h"
#include "PosWCS.h"
#include "CoordPair.h"

#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "slatec.h"
#include "ExceptionErrorHandler.h"


/* Method: Poscorr3xmm() - Constructor */
Poscorr3xmm::Poscorr3xmm()
{
///  PROC;

  /* Get the parameters from this command line */
    string srcfile = stringParameter("srclist");

  /* Open the source list */
    DataSet * srclist = dataSetServer->open( srcfile, DataSet::Modify);

  /* Check whether this file has already been corrected */
    if (srclist->hasAttribute("ELLBCORR"))
       if (srclist->attribute("ELLBCORR")->asBool())
          errstr<<"Source list "<<srcfile<<" has already been corrected"
                <<Sas::error("AlreadyCorrected");

  /* Open the source table */
    Table * srctable = srclist->table("SRCLIST");

   /* Get the RA, DEC of each source */
    Column *cra = srctable->column("RA");
    Column *cdec = srctable->column("DEC");
    ColumnData *cdra = cra->data();
    ColumnData *cddec = cdec->data();
    double *ra = cdra->real64Data();
    double *dec = cddec->real64Data();

  /* Set up a position utilities object - using EPIC-pn to define
     the coordinate frame */
    PositionUtils *putils = new PositionUtils(Xmm::toInstrument("EPN"));
    
  /* Set up a poswcs object */
         /* Pointing */
    double rapnt  = srclist->attribute("RA_PNT")->asReal();
    double decpnt  = srclist->attribute("DEC_PNT")->asReal();
    double papnt  = srclist->attribute("PA_PNT")->asReal();

         // Ignore warning due to missing REF keywords
    const PosWCS *poswcs=0;
//    ExceptionErrorHandler local;
  //  ErrorHandler* oldHandler = errHandler.handler(&local);
    //try {
        poswcs =   new PosWCS(srclist, rapnt, decpnt, papnt, false);
  //  } catch(...) { // Do nothing
    //}
 //   errHandler.handler(oldHandler);

  // Transform to radians
    double dtor = M_PI / 180.0;
    double rad_rapnt = rapnt * dtor;
    double rad_decpnt = decpnt * dtor;

  // Get the date 
    string dateobs = srclist->attribute("DATE-OBS")->asString();

  /* Loop over sources */
    for (unsigned i=0; i<srctable->rows(); ++i)
     {
      // Convert celestial position to theta, phi
       SourcePosition s = putils->sourcePosition("EQPOS", "", dateobs, 
                           ra[i], dec[i], poswcs);

      // Get off-axis angle in arcseconds
       double theta_arcsec = s.theta();

      // Calculate the source centre offset 
       double theta_offset = thetaOffset(theta_arcsec);  
       double theta_ratio = (theta_arcsec - theta_offset) / theta_arcsec;

      // Convert ra, dec to radians
       double raIn = ra[i] * dtor; 
       double decIn = dec[i] * dtor;

      // Calculate the new position after conversion
       CoordPair newcel = calcnewpos(rad_rapnt, rad_decpnt,  
                                     raIn, decIn, theta_ratio);

      // Set new value of RA and DEC in degrees
       ra[i] = newcel.x() / dtor;  
       dec[i] = newcel.y()/ dtor;

      // Give debug
       errstr<<"RA, DEC (in / out):"<<raIn/dtor<<" "<<ra[i]<<" "
             <<decIn/dtor<<" "<<dec[i]<<" theta off:"<<theta_offset
             <<std::setprecision(9)
             << Sas::message(Msg::App, Msg::Verbose);
     }

    delete poswcs;

  // NB: This doesnt update galactic keywords - it probably should

  /* Add keyword to this file */
    if (srclist->hasAttribute("ELLBCORR"))
       *srclist->attribute("ELLBCORR") = true;
    else 
       *srclist->addAttribute("ELLBCORR", 
                  "Has ELLBETA centre correction been applied") = true;
 
  // Close up and go home
    delete cdra;
    delete cddec;
    dataSetServer->close(srclist);
}

Poscorr3xmm::~Poscorr3xmm()
{}

CoordPair Poscorr3xmm::calcnewpos(double ra, double dec, double thetaoffset)
  { 

return CoordPair((ra - thetaoffset), dec);
}

/* Method to calculate a new celestial position 
   from stuff */
CoordPair Poscorr3xmm::calcnewpos(double ra_pnt, double dec_pnt,
     double ra, double dec, double theta_ratio)
{
  // Set DEC reference  
   double dec_ref=0.0;

   if(dec_pnt > 0.0) 
      dec_ref = dec_pnt - 0.5 * M_PI;
   else
      dec_ref = dec_pnt + 0.5 * M_PI;

  // Use these two 3-vectors to compute coordinate rotation matrix and inverse
   vector<double> vpnt = cona2v(ra_pnt, dec_pnt);
   vector<double> vref = cona2v(ra_pnt, dec_ref);

   vector<vector<double> > ctos, stoc;
   ctos = congen(vpnt, vref);
   stoc = conmin(ctos);

  // Get vector for the source position
   vector<double> vx = cona2v(ra, dec);
   vector<double> newvx = rotvec(vx, ctos);

  // Modify new vector by the offset ratio
   newvx[0] = newvx[0] * theta_ratio;
   newvx[1] = newvx[1] * theta_ratio;
   newvx[2] = sqrt(1.0 - newvx[0] * newvx[0] - newvx[1] * newvx[1]);

  // Rotate back to celestial frame
   vector<double> newcel = rotvec(newvx, stoc);

   double raout = atan2(newcel[1], newcel[0]);
   if (raout < 0.0) raout += M_PI * 2.0; 
   double decout = asin(newcel[2]); 
  

   return CoordPair(raout, decout); 
}

// congen
vector<vector<double> > Poscorr3xmm::congen(vector<double> vz, vector<double> vx)
{
  // Declare a matrix
   vector<vector<double> > m(3, std::vector<double>(3,0.0));

   for (int i=0; i<3; i++) 
       m[i][2] = vz[i];

   vector<double> vec3(3), vec3b(3);

  // Find cross-product of input vectors
   vec3 = convcp(vz, vx);
   
  // Set result in matrix
   for (int i=0; i<3; i++)
       m[i][1] = vec3[i];

  // Find cross-product of other vectors
   vec3b = convcp(vec3, vz);

  // Update the output matrix  
   for (int i=0; i<3; i++)
       m[i][0] = vec3b[i];   

   return m;
}

// convcp
vector<double> Poscorr3xmm::convcp(vector<double> va, vector<double> vb)
{
   vector<double> vc(3);

   vc[0] = va[1]*vb[2] - va[2]*vb[1];
   vc[1] = va[2]*vb[0] - va[0]*vb[2];
   vc[2] = va[0]*vb[1] - va[1]*vb[0];

   double vnorm = sqrt(vc[0]*vc[0] + vc[1]*vc[1] + vc[2]*vc[2]);
   
   for (int i=0; i<3; i++)
        vc[i] = vc[i] / vnorm;

return vc;
}
   
// conmin
// Transposes a matrix
vector<vector<double> > Poscorr3xmm::conmin(vector<vector<double> > m_in)
{
   vector<vector<double> > m_out(3, std::vector<double>(3,0.0));

   for (int i=0; i<3; i++)
     for (int j=0; j<3; j++)
         m_out[i][j] = m_in[j][i];

   return m_out;
}

//
vector<double> Poscorr3xmm::cona2v(double ra, double dec)
{
   vector<double> v;

   v.push_back(cos(ra) * cos(dec));
   v.push_back(sin(ra) * cos(dec));
   v.push_back(sin(dec));

   return v;
}

// Rotate a vector
vector<double> Poscorr3xmm::rotvec(vector<double> vin, vector<vector<double> > m)
{
   vector<double> vout(3,0.0);

   for (int j=0; j<3; j++)
     for (int i=0; i<3; i++) {
        vout[j] += vin[i] * m[i][j];
     }

   return vout;
}


/* Method to calculate the source centre offset in arcsecs
   Input: thetaRad - off-axis angle in arcseconds */
double Poscorr3xmm::thetaOffset(double thetaAsec)
{
  PROC;

   // Hard code coefficients here - these have been derived for the 
   // all-EPIC images with all of the approximations that this implies
    double off0 = 0.0158736;
    double off1 = -0.036454;
    double off2 = 0.00942358;
    double off3 = -0.000277756;

   // Calculate the offset using a cubic fit derived by Andy Read, 2012 Dec 14
   // Theta should be in arcminutes
    //double theta = thetaAsec * 180.0 * 60.0 / M_PI;
    double theta = thetaAsec / 60.0;

    double offset = off0 + off1 * theta + off2 * theta * theta +
                    off3 * theta * theta * theta;

    return offset;
}
 
